##
# Copyright (C) Microsoft Corporation. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause-Patent
#
# Python script that converts a raw IVRS table into a struct, the spec version is based on
# https://www.amd.com/system/files/TechDocs/48882_IOMMU.pdf
##
"""Module for converting a raw IVRS table into a struct."""

import struct
import sys
import xml.etree.ElementTree as ET
from enum import IntEnum
from typing import Optional

IVRSParserVersion = "1.00"
# spell-checker:ignore IVMD, IOMMUEFR


class IVRS_TABLE(object):
    """Object representing a IVRS Table."""

    def __init__(self, data: Optional[bytes] = None) -> "IVRS_TABLE":
        """Inits an empty object."""
        self.acpi_header = None
        self.SubStructs = list()
        self.IVMD_list = list()

        if data is not None:
            self.Decode(data)

    def Decode(self, data: bytes) -> None:
        """Loads data into the object from the header_byte_array."""
        self.acpi_header = IVRS_TABLE.ACPI_TABLE_HEADER(data[: IVRS_TABLE.ACPI_TABLE_HEADER.struct_format_size])

        # Start from the end of ACPI header, but store the parsed length for verification
        t_length = self.acpi_header.Length
        self.acpi_header.Length = IVRS_TABLE.ACPI_TABLE_HEADER.struct_format_size
        t_data = data[IVRS_TABLE.ACPI_TABLE_HEADER.struct_format_size :]

        # sanity check on incoming data
        Checksum8 = IVRS_TABLE.validateChecksum8(data)
        if Checksum8 != 0:
            raise Exception(
                "Incoming data checksum does not add up: checksum field %x, calculated is %x"
                % (self.acpi_header.Checksum, Checksum8)
            )

        while len(t_data) > 0:
            # Get type and length of remapping struct
            remapping_header = self.REMAPPING_STRUCT_HEADER(t_data)

            # Parse remapping struct
            if (
                (remapping_header.Type == IVRS_TABLE.IVHD_STRUCT.IVHD_TYPE.TYPE_10H)
                or (remapping_header.Type == IVRS_TABLE.IVHD_STRUCT.IVHD_TYPE.TYPE_11H)
                or (remapping_header.Type == IVRS_TABLE.IVHD_STRUCT.IVHD_TYPE.TYPE_40H)
            ):
                remapping_header = self.IVHD_STRUCT(t_data)
                self.addIVHDEntry(remapping_header)
            elif (
                (remapping_header.Type == IVRS_TABLE.IVMD_STRUCT.IVMD_TYPE.TYPE_20H)
                or (remapping_header.Type == IVRS_TABLE.IVMD_STRUCT.IVMD_TYPE.TYPE_21H)
                or (remapping_header.Type == IVRS_TABLE.IVMD_STRUCT.IVMD_TYPE.TYPE_22H)
            ):
                remapping_header = self.IVMD_STRUCT(t_data[: IVRS_TABLE.IVMD_STRUCT.struct_format_size])
                self.addIVMDEntry(remapping_header)
                if remapping_header.Type == IVRS_TABLE.IVMD_STRUCT.IVMD_TYPE.TYPE_20H:
                    self.IVRSBit = 0
            else:
                print("Reserved remapping struct found in IVRS table %d" % remapping_header.Type)
                sys.exit(-1)

            # Update data position
            t_data = t_data[remapping_header.Length :]

        if (self.acpi_header.Length != t_length) or (len(t_data) != 0):
            raise Exception(
                "IVRS length does not add up. Parsed len: %d, reported len: %d" % (t_length, self.acpi_header.Length)
            )

    def Encode(self) -> bytes:
        r"""Serializes the object.

        Returns:
            (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
        """
        bytes_str = b""

        # Append ACPI header
        bytes_str += self.acpi_header.Encode()

        # All IVHD/IVMD entries
        for ivxd in self.SubStructs:
            bytes_str += ivxd.Encode()
        return bytes_str

    def ToXmlElementTree(self) -> ET.Element:
        """Transforms the object into an xml item and returns it."""
        root = ET.Element("IVRSTable")
        root.append(self.acpi_header.ToXmlElementTree())
        for sub in self.SubStructs:
            root.append(sub.ToXmlElementTree())

        return root

    def DumpInfo(self) -> None:
        """Prints internal information to the console."""
        self.acpi_header.DumpInfo()

        for sub in self.SubStructs:
            sub.DumpInfo()

    @staticmethod
    def validateChecksum8(data: bytes) -> bool:
        """Validate the Checksum."""
        return sum(data) & 0xFF

    def updateACPISum(self) -> None:
        """Update the ACPI Checksum."""
        temp_sum = 0
        # Clear the checksum before calculating sum
        self.acpi_header.Checksum = 0
        temp_str = self.Encode()
        temp_sum = sum(temp_str)
        self.acpi_header.Checksum = (0x100 - (temp_sum & 0xFF)) & 0xFF

    def addIVHDEntry(self, ivhd: "IVRS_TABLE.IVHD_STRUCT") -> None:
        """Append entry to the list, update length and checksum."""
        self.acpi_header.Length += len(ivhd.Encode())
        self.SubStructs.append(ivhd)
        self.updateACPISum()

    def addIVMDEntry(self, ivmd: "IVRS_TABLE.IVMD_STRUCT") -> None:
        """Append entry to the list, update length and checksum."""
        self.acpi_header.Length += len(ivmd.Encode())
        # IVMD has to follow the corresponding IVHD, thus the list records all entries to maintain order
        self.SubStructs.append(ivmd)
        self.IVMD_list.append(ivmd)
        self.updateACPISum()

    def IVRSBitEnabled(self) -> bool:
        """Returns the IVRSBit Value."""
        return bool(self.acpi_header.IVRSBit)

    class ACPI_TABLE_HEADER(object):
        """Object representing a ACPI Table Header."""

        struct_format = "=4sIBB6s8sI4sIIQ"
        struct_format_size = struct.calcsize(struct_format)

        def __init__(self, data: Optional[bytes] = None) -> "IVRS_TABLE.ACPI_TABLE_HEADER":
            """Inits an empty object."""
            self.Signature = None
            self.Length = 0
            self.Revision = 0
            self.Checksum = 0
            self.OEMID = 0
            self.OEMTableID = 0
            self.OEMRevision = 0
            self.CreatorID = 0
            self.CreatorRevision = 0
            self.IVinfo = None
            self.Reserved = 0

            self.IVRSBit = 0

            if data is not None:
                self.Decode(data)

        def Decode(self, header_byte_array: bytes) -> None:
            """Loads data into the object from the header_byte_array."""
            (
                self.Signature,
                self.Length,
                self.Revision,
                self.Checksum,
                self.OEMID,
                self.OEMTableID,
                self.OEMRevision,
                self.CreatorID,
                self.CreatorRevision,
                self.IVinfo,
                self.Reserved,
            ) = struct.unpack(IVRS_TABLE.ACPI_TABLE_HEADER.struct_format, header_byte_array)

            self.IVRSBit = self.IVinfo & 0x02
            if (self.IVinfo & 0x1E) == 0:
                sys.exit(-1)

        def Encode(self) -> bytes:
            r"""Serializes the object.

            Returns:
                (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
            """
            return struct.pack(
                self.struct_format,
                self.Signature,
                self.Length,
                self.Revision,
                self.Checksum,
                self.OEMID,
                self.OEMTableID,
                self.OEMRevision,
                self.CreatorID,
                self.CreatorRevision,
                self.IVinfo,
                self.Reserved,
            )

        def DumpInfo(self) -> None:
            """Prints internal information to the console."""
            print("  ACPI Table Header")
            print("------------------------------------------------------------------")
            print("Signature          : {Signature:s}".format(Signature=self.Signature.decode()))
            print("Length             : 0x{Length:08X}".format(Length=self.Length))
            print("Revision           : 0x{Revision:02X}".format(Revision=self.Revision))
            print("Checksum           : 0x{Checksum:02X}".format(Checksum=self.Checksum))
            print("OEM ID             : {OEMID:s}".format(OEMID=self.OEMID.decode()))
            print("OEM Table ID       : {OEMTableID:s}".format(OEMTableID=self.OEMTableID.decode()))
            print("OEM Revision       : 0x{OEMRevision:08X}".format(OEMRevision=self.OEMRevision))
            print("Creator ID         : {CreatorID:s}".format(CreatorID=self.CreatorID.decode()))
            print("Creator Revision   : 0x{CreatorRevision:08X}".format(CreatorRevision=self.CreatorRevision))
            print("IVinfo             : 0x{IVinfo:08X}".format(IVinfo=self.IVinfo))

        def ToXmlElementTree(self) -> ET.Element:
            """Transforms the object into an xml item and returns it."""
            xml_repr = ET.Element("AcpiTableHeader")
            xml_repr.set("Signature", "%s" % self.Signature)
            xml_repr.set("Length", "0x%X" % self.Length)
            xml_repr.set("Revision", "0x%X" % self.Revision)
            xml_repr.set("Checksum", "0x%X" % self.Checksum)
            xml_repr.set("OEMID", "%s" % self.OEMID)
            xml_repr.set("OEMTableID", "%s" % self.OEMTableID)
            xml_repr.set("OEMRevision", "0x%X" % self.OEMRevision)
            xml_repr.set("CreatorID", "%s" % self.CreatorID)
            xml_repr.set("CreatorRevision", "0x%X" % self.CreatorRevision)
            xml_repr.set("IVinfo", "0x%X" % self.IVinfo)
            return xml_repr

    class REMAPPING_STRUCT_HEADER(object):
        """Generic ReMapping Struct Header."""

        struct_format = "=B"
        struct_format_size = struct.calcsize(struct_format)

        def __init__(self, header_byte_array: bytes) -> "IVRS_TABLE.REMAPPING_STRUCT_HEADER":
            """Inits an empty object."""
            (self.Type,) = struct.unpack(
                IVRS_TABLE.REMAPPING_STRUCT_HEADER.struct_format,
                header_byte_array[: IVRS_TABLE.REMAPPING_STRUCT_HEADER.struct_format_size],
            )

    class IVHD_STRUCT(REMAPPING_STRUCT_HEADER):
        """Object representing a IVHD Struct."""

        # cspell:disable-next disable the spell checker from thinking this a word
        struct_format = "=BBHHHQHHI"
        struct_format_size = struct.calcsize(struct_format)
        ex_format = "=QQ"
        ex_format_size = struct.calcsize(ex_format)

        class IVHD_TYPE(IntEnum):
            """IVMD_STRUCT type enum."""

            TYPE_10H = 0x10
            TYPE_11H = 0x11
            TYPE_40H = 0x40

        def __init__(self, data: Optional[bytes] = None) -> "IVRS_TABLE.IVHD_STRUCT.IVHD_TYPE":
            """Inits an empty object."""
            self.Type = None
            self.Flags = None
            self.Length = 0
            self.DeviceID = 0
            self.CapabilityOffset = 0
            self.IOMMUBaseAddress = 0
            self.SegmentGroup = 0
            self.IOMMUInfo = None
            self.IOMMUFeatureInfo = None
            self.IOMMUEFRImage = None
            self.Reserved = 0

            self.DeviceTableEntries = list()

            if data is not None:
                self.Decode(data)

        def Decode(self, header_byte_array: bytes) -> None:
            """Loads data into the object from the header_byte_array."""
            (
                self.Type,
                self.Flags,
                t_Length,
                self.DeviceID,
                self.CapabilityOffset,
                self.IOMMUBaseAddress,
                self.SegmentGroup,
                self.IOMMUInfo,
                self.IOMMUFeatureInfo,
            ) = struct.unpack(
                IVRS_TABLE.IVHD_STRUCT.struct_format, header_byte_array[: IVRS_TABLE.IVHD_STRUCT.struct_format_size]
            )

            self.Length = 0

            if (self.Type == IVRS_TABLE.IVHD_STRUCT.IVHD_TYPE.TYPE_11H) or (
                self.Type == IVRS_TABLE.IVHD_STRUCT.IVHD_TYPE.TYPE_40H
            ):
                ivhd_hdr_size = IVRS_TABLE.IVHD_STRUCT.struct_format_size + IVRS_TABLE.IVHD_STRUCT.ex_format_size
                (self.IOMMUEFRImage, self.Reserved) = struct.unpack(
                    IVRS_TABLE.IVHD_STRUCT.ex_format,
                    header_byte_array[IVRS_TABLE.IVHD_STRUCT.struct_format_size : ivhd_hdr_size],
                )
            else:
                ivhd_hdr_size = IVRS_TABLE.IVHD_STRUCT.struct_format_size

            header_byte_array = header_byte_array[ivhd_hdr_size:]
            bytes_left = t_Length - ivhd_hdr_size
            self.Length += ivhd_hdr_size

            # Get Sub Structs
            while bytes_left > 0:
                device_scope = IVRS_TABLE.DEVICE_TABLE_ENTRY.Factory(header_byte_array)
                header_byte_array = header_byte_array[device_scope.Length :]
                bytes_left -= device_scope.Length
                if device_scope.Type != IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.RANGE_END:
                    self.addDTEEntry(device_scope)

            if (t_Length != self.Length) or (bytes_left != 0):
                raise Exception(
                    "IVHD length does not add up. Parsed len: %d, reported len: %d" % (self.Length, t_Length)
                )

        def Encode(self) -> bytes:
            r"""Serializes the object.

            Returns:
                (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
            """
            byte_str = b""
            byte_str += struct.pack(
                IVRS_TABLE.IVHD_STRUCT.struct_format,
                self.Type,
                self.Flags,
                self.Length,
                self.DeviceID,
                self.CapabilityOffset,
                self.IOMMUBaseAddress,
                self.SegmentGroup,
                self.IOMMUInfo,
                self.IOMMUFeatureInfo,
            )

            if self.IOMMUEFRImage is not None:
                byte_str += struct.pack(IVRS_TABLE.IVHD_STRUCT.ex_format, self.IOMMUEFRImage, self.Reserved)

            for dte in self.DeviceTableEntries:
                byte_str += dte.Encode()

            return byte_str

        def addDTEEntry(self, dte: "IVRS_TABLE.DEVICE_TABLE_ENTRY") -> None:
            """Append raw data, update length. checksum will be left untouched."""
            self.Length += len(dte.Encode())
            self.DeviceTableEntries.append(dte)

        def ToXmlElementTree(self) -> ET.Element:
            """Transforms the object into an xml item and returns it."""
            xml_repr = ET.Element("IVHD")
            xml_repr.set("Type", "0x%X" % self.Type)
            xml_repr.set("Flags", "0x%X" % self.Flags)
            xml_repr.set("Length", "0x%X" % self.Length)
            xml_repr.set("IOMMUDeviceID", "0x%X" % self.DeviceID)
            xml_repr.set("CapabilityOffset", "0x%X" % self.CapabilityOffset)
            xml_repr.set("IOMMUBaseAddress", "0x%X" % self.IOMMUBaseAddress)
            xml_repr.set("SegmentGroup", "0x%X" % self.SegmentGroup)
            xml_repr.set("IOMMUInfo", "0x%X" % self.IOMMUInfo)
            xml_repr.set("IOMMUFeatureInfo", "0x%X" % self.IOMMUFeatureInfo)

            if (self.Type == IVRS_TABLE.IVHD_STRUCT.IVHD_TYPE.TYPE_11H) or (
                self.Type == IVRS_TABLE.IVHD_STRUCT.IVHD_TYPE.TYPE_40H
            ):
                xml_repr.set("IOMMUEFRImage", "0x%X" % self.IOMMUEFRImage)

            # Add SubStructs
            for item in self.DeviceTableEntries:
                xml_repr.append(item.ToXmlElementTree())

            return xml_repr

        def DumpInfo(self) -> None:
            """Prints internal information to the console."""
            print("\t  IVHD")
            print("\t----------------------------------------------------------------")
            print("\tType                  : 0x{Type:02X}".format(Type=self.Type))
            print("\tFlags                 : 0x{Flags:02X}".format(Flags=self.Flags))
            print("\tLength                : 0x{Length:04X}".format(Length=self.Length))
            print("\tIOMMU Device ID       : 0x{DeviceID:04X}".format(DeviceID=self.DeviceID))
            print("\tCapability Offset     : 0x{CapabilityOffset:04X}".format(CapabilityOffset=self.CapabilityOffset))
            print("\tIOMMU Base Address    : 0x{IOMMUBaseAddress:016X}".format(IOMMUBaseAddress=self.IOMMUBaseAddress))
            print("\tSegment Group         : 0x{SegmentGroup:04X}".format(SegmentGroup=self.SegmentGroup))
            print("\tIOMMU Info            : 0x{IOMMUInfo:04X}".format(IOMMUInfo=self.IOMMUInfo))
            print("\tIOMMU Feature Info    : 0x{IOMMUFeatureInfo:08X}".format(IOMMUFeatureInfo=self.IOMMUFeatureInfo))

            for item in self.DeviceTableEntries:
                item.DumpInfo()

    class IVMD_STRUCT(REMAPPING_STRUCT_HEADER):
        """Object representing an IVMD Struct."""

        # cspell:disable-next disable spell checker from thinking this is a word
        struct_format = "=BBHHHQQQ"
        struct_format_size = struct.calcsize(struct_format)

        class IVMD_TYPE(IntEnum):
            """IVMD_STRUCT type enum."""

            TYPE_20H = 0x20  # All peripherals
            TYPE_21H = 0x21  # Specified peripheral
            TYPE_22H = 0x22  # Peripheral range

        def __init__(self, data: Optional[bytes] = None) -> "IVRS_TABLE.IVMD_STRUCT":
            """Inits an empty object."""
            self.Type = None
            self.Flags = None
            self.Length = 0
            self.DeviceID = 0
            self.AuxiliaryData = None
            self.Reserved = 0
            self.IVMDStartAddress = 0
            self.IVMDMemoryBlockLength = 0

            if data is not None:
                self.Decode(data)

        def Decode(self, header_byte_array: bytes) -> None:
            """Loads data into the object from the header_byte_array."""
            (
                self.Type,
                self.Flags,
                self.Length,
                self.DeviceID,
                self.AuxiliaryData,
                self.Reserved,
                self.IVMDStartAddress,
                self.IVMDMemoryBlockLength,
            ) = struct.unpack(IVRS_TABLE.IVMD_STRUCT.struct_format, header_byte_array)
            # IVMD is simple, the length is fixed, so assert if not
            if self.Length != len(header_byte_array):
                raise Exception("Bad IVMD entry size %d, expecting %d" % (self.Length, len(header_byte_array)))

        def Encode(self) -> bytes:
            r"""Serializes the object.

            Returns:
                (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
            """
            return struct.pack(
                IVRS_TABLE.IVMD_STRUCT.struct_format,
                self.Type,
                self.Flags,
                self.Length,
                self.DeviceID,
                self.AuxiliaryData,
                self.Reserved,
                self.IVMDStartAddress,
                self.IVMDMemoryBlockLength,
            )

        def ToXmlElementTree(self) -> ET.Element:
            """Transforms the object into an xml item and returns it."""
            xml_repr = ET.Element("IVMD")

            xml_repr.set("Type", "0x%X" % self.Type)
            xml_repr.set("Flags", "0x%X" % self.Flags)
            xml_repr.set("Length", "0x%X" % self.Length)
            xml_repr.set("DeviceID", "0x%X" % self.DeviceID)
            if self.Type != IVRS_TABLE.IVMD_STRUCT.IVMD_TYPE.TYPE_22H:
                xml_repr.set("AuxiliaryData", "0x%X" % self.AuxiliaryData)
            else:
                xml_repr.set("EndofRange", "0x%X" % self.AuxiliaryData)
            xml_repr.set("Reserved", "0x%X" % self.Reserved)
            xml_repr.set("IVMDStartAddress", "0x%X" % self.IVMDStartAddress)
            xml_repr.set("IVMDMemoryBlockLength", "0x%X" % self.IVMDMemoryBlockLength)

            return xml_repr

        def DumpInfo(self) -> None:
            """Prints internal information to the console."""
            print("\t  IVMD")
            print("\t----------------------------------------------------------------")
            print("\tType                                 : 0x{Type:02X}".format(Type=self.Type))
            print("\tFlags                                : 0x{Flags:02X}".format(Flags=self.Flags))
            print("\tLength                               : 0x{Length:04X}".format(Length=self.Length))
            print("\tDeviceID                             : 0x{DeviceID:04X}".format(DeviceID=self.DeviceID))
            print(
                "\tAuxiliaryData                        : 0x{AuxiliaryData:04X}".format(
                    AuxiliaryData=self.AuxiliaryData
                )
            )
            print("\tReserved                             : 0x{Reserved:016X}".format(Reserved=self.Reserved))
            print(
                "\tIVMD Start Address                   : 0x{IVMDStartAddress:016X}".format(
                    IVMDStartAddress=self.IVMDStartAddress
                )
            )
            print(
                "\tIVMD Memory Block Length             : 0x{IVMDMemoryBlockLength:016X}".format(
                    IVMDMemoryBlockLength=self.Type
                )
            )

    class DEVICE_TABLE_ENTRY(object):
        """A Generic Device Table Entry."""

        struct_format = "=BHB"
        struct_format_size = struct.calcsize(struct_format)
        dte_var_ext_format = "=8s8sBB"
        dte_var_len = struct_format_size + struct.calcsize(dte_var_ext_format)

        class DTE_TYPE(IntEnum):
            """Enum for Device Table Entry Types."""

            RESERVED = 0
            ALL = 1
            SELECT = 2
            RANGE_START = 3
            RANGE_END = 4
            ALIAS_SELECT = 66
            ALIAS_RANGE_START = 67
            EX_SELECT = 70
            EX_RANGE_START = 71
            SPECIAL = 72
            ACPI = 240

        @staticmethod
        def Factory(data: bytes) -> Optional["IVRS_TABLE.DEVICE_TABLE_ENTRY_SPECIAL"]:
            """Factory method to generate a specific table entry type."""
            if data is None:
                raise Exception("Invalid File stream")

            RemapHeader = IVRS_TABLE.REMAPPING_STRUCT_HEADER(data)
            Type = RemapHeader.Type

            if Type == IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.RESERVED:
                return IVRS_TABLE.DEVICE_TABLE_ENTRY_RESERVED(data)
            elif Type == IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.ALL:
                return IVRS_TABLE.DEVICE_TABLE_ENTRY_ALL(data)
            elif Type == IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.SELECT:
                return IVRS_TABLE.DEVICE_TABLE_ENTRY_SELECT(data)
            elif Type == IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.RANGE_START:
                return IVRS_TABLE.DEVICE_TABLE_ENTRY_RANGE_START(data)
            elif Type == IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.ALIAS_SELECT:
                return IVRS_TABLE.DEVICE_TABLE_ENTRY_ALIAS_SELECT(data)
            elif Type == IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.ALIAS_RANGE_START:
                return IVRS_TABLE.DEVICE_TABLE_ENTRY_ALIAS_RANGE_START(data)
            elif Type == IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.EX_SELECT:
                return IVRS_TABLE.DEVICE_TABLE_ENTRY_EX_SELECT(data)
            elif Type == IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.EX_RANGE_START:
                return IVRS_TABLE.DEVICE_TABLE_ENTRY_EX_RANGE_START(data)
            elif Type == IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.SPECIAL:
                return IVRS_TABLE.DEVICE_TABLE_ENTRY_SPECIAL(data)
            elif Type == IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.ACPI:
                return IVRS_TABLE.DEVICE_TABLE_ENTRY_ACPI(data)
            else:
                return None

    class DEVICE_TABLE_ENTRY_RESERVED(object):
        """Object representing a Device Table Entry RESERVED."""

        def __init__(self, data: Optional[bytes] = None) -> "IVRS_TABLE.DEVICE_TABLE_ENTRY":
            """Inits an empty object."""
            self.Type = 0
            self.DeviceID = 0
            self.DTESetting = 0
            self.TypeString = None
            self.Length = 0
            if data is not None:
                self.Decode(data)

        def Decode(self, header_byte_array: bytes) -> None:
            """Loads data into the object from the header_byte_array."""
            (self.Type, self.DeviceID, self.DTESetting) = struct.unpack(
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                header_byte_array[: IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size],
            )
            if self.Type != IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.RESERVED:
                raise Exception(
                    "Input device type (%d) does not match expectation (%d)",
                    IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.RESERVED,
                    self.Type,
                )

            self.TypeString = "Reserved"
            self.Length = IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size

        def Encode(self) -> bytes:
            r"""Serializes the object.

            Returns:
                (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
            """
            return struct.pack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format, self.Type, self.DeviceID, self.DTESetting)

        def ToXmlElementTree(self) -> ET.Element:
            """Transforms the object into an xml item and returns it."""
            xml_item = ET.Element(self.TypeString.replace(" ", ""))
            xml_item.set("DeviceID", "0x%X" % (self.DeviceID))
            xml_item.set("DTESetting", "0x%X" % (self.DTESetting))
            xml_item.set("Type", "0x%X" % self.Type)
            return xml_item

        def DumpInfo(self) -> None:
            """Prints internal information to the console."""
            print("\t\t  {TypeString:s}".format(TypeString=self.TypeString))
            print("\t\t--------------------------------------------------")
            print("\t\tType                  : 0x{Type:02X}".format(Type=self.Type))
            print("\t\tDevice ID             : 0x{DeviceID:04X}".format(DeviceID=self.DeviceID))
            print("\t\tDTE Setting           : 0x{DTESetting:02X}".format(DTESetting=self.DTESetting))

    class DEVICE_TABLE_ENTRY_ALL(object):
        """Object representing a Device Table Entry ALL."""

        def __init__(self, data: Optional[bytes] = None) -> "IVRS_TABLE.DEVICE_TABLE_ENTRY_ALL":
            """Inits an empty object."""
            self.Type = 0
            self.DeviceID = 0
            self.DTESetting = 0
            self.TypeString = None
            self.Length = 0
            if data is not None:
                self.Decode(data)

        def Decode(self, header_byte_array: bytes) -> None:
            """Loads data into the object from the header_byte_array."""
            (self.Type, self.DeviceID, self.DTESetting) = struct.unpack(
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                header_byte_array[: IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size],
            )
            if self.Type != IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.ALL:
                raise Exception(
                    "Input device type (%d) does not match expectation (%d)",
                    IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.ALL,
                    self.Type,
                )

            self.TypeString = "All"
            self.Length = IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size

        def Encode(self) -> bytes:
            r"""Serializes the object.

            Returns:
                (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
            """
            return struct.pack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format, self.Type, self.DeviceID, self.DTESetting)

        def ToXmlElementTree(self) -> ET.Element:
            """Transforms the object into an xml item and returns it."""
            xml_item = ET.Element(self.TypeString.replace(" ", ""))
            xml_item.set("Type", "0x%X" % self.Type)
            xml_item.set("DeviceID", "0x%X" % (self.DeviceID))
            xml_item.set("DTESetting", "0x%X" % (self.DTESetting))
            return xml_item

        def DumpInfo(self) -> None:
            """Prints internal information to the console."""
            print("\t\t  {TypeString:s}".format(TypeString=self.TypeString))
            print("\t\t--------------------------------------------------")
            print("\t\tType                  : 0x{Type:02X}".format(Type=self.Type))
            print("\t\tDevice ID             : 0x{DeviceID:04X}".format(DeviceID=self.DeviceID))
            print("\t\tDTE Setting           : 0x{DTESetting:02X}".format(DTESetting=self.DTESetting))

    class DEVICE_TABLE_ENTRY_SELECT(object):
        """Object representing a Device Table Entry SELECT."""

        def __init__(self, data: Optional[bytes] = None) -> "IVRS_TABLE.DEVICE_TABLE_ENTRY_SELECT":
            """Inits an empty object."""
            self.Type = 0
            self.DeviceID = 0
            self.DTESetting = 0
            self.TypeString = None
            self.Length = 0
            if data is not None:
                self.Decode(data)

        def Decode(self, header_byte_array: bytes) -> None:
            """Loads data into the object from the header_byte_array."""
            (self.Type, self.DeviceID, self.DTESetting) = struct.unpack(
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                header_byte_array[: IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size],
            )
            if self.Type != IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.SELECT:
                raise Exception(
                    "Input device type (%d) does not match expectation (%d)",
                    IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.SELECT,
                    self.Type,
                )

            self.TypeString = "Reserved"
            self.Length = IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size

        def Encode(self) -> bytes:
            r"""Serializes the object.

            Returns:
                (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
            """
            return struct.pack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format, self.Type, self.DeviceID, self.DTESetting)

        def ToXmlElementTree(self) -> ET.Element:
            """Transforms the object into an xml item and returns it."""
            xml_item = ET.Element(self.TypeString.replace(" ", ""))
            xml_item.set("Type", "0x%X" % self.Type)
            xml_item.set("DeviceID", "0x%X" % (self.DeviceID))
            xml_item.set("DTESetting", "0x%X" % (self.DTESetting))
            return xml_item

        def DumpInfo(self) -> None:
            """Prints internal information to the console."""
            print("\t\t  {TypeString:s}".format(TypeString=self.TypeString))
            print("\t\t--------------------------------------------------")
            print("\t\tType                  : 0x{Type:02X}".format(Type=self.Type))
            print("\t\tDevice ID             : 0x{DeviceID:04X}".format(DeviceID=self.DeviceID))
            print("\t\tDTE Setting           : 0x{DTESetting:02X}".format(DTESetting=self.DTESetting))

    class DEVICE_TABLE_ENTRY_RANGE_START(object):
        """Object representing a Device Table Entry RANGE_START."""

        def __init__(self, data: Optional[bytes] = None) -> "IVRS_TABLE.DEVICE_TABLE_ENTRY_RANGE_START":
            """Inits an empty object."""
            self.Type = 0
            self.DeviceID = 0
            self.DTESetting = 0
            self.TypeString = None
            self.Length = 0
            if data is not None:
                self.Decode(data)

        def Decode(self, header_byte_array: bytes) -> None:
            """Loads data into the object from the header_byte_array."""
            (self.Type, self.DeviceID, self.DTESetting) = struct.unpack(
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                header_byte_array[: IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size],
            )
            if self.Type != IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.RANGE_START:
                raise Exception(
                    "Input device type (%d) does not match expectation (%d)",
                    IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.RANGE_START,
                    self.Type,
                )

            self.TypeString = "Range"
            self.Length = IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size
            (Type, self.EndDeviceID, _) = struct.unpack(
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                header_byte_array[self.Length : (self.Length + IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size)],
            )
            if Type != IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.RANGE_END:
                print("Start of range does not follow end of range")
                sys.exit(-1)
            self.Length += IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size

        def Encode(self) -> bytes:
            r"""Serializes the object.

            Returns:
                (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
            """
            byte_str = struct.pack(
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format, self.Type, self.DeviceID, self.DTESetting
            )
            byte_str += struct.pack(
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.RANGE_END,
                self.EndDeviceID,
                0,
            )
            return byte_str

        def ToXmlElementTree(self) -> ET.Element:
            """Transforms the object into an xml item and returns it."""
            xml_item = ET.Element(self.TypeString.replace(" ", ""))
            xml_item.set("Type", "0x%X" % self.Type)
            xml_item.set("StartofRange", "0x%X" % self.DeviceID)
            xml_item.set("EndofRange", "0x%X" % (self.EndDeviceID))
            xml_item.set("DTESetting", "0x%X" % (self.DTESetting))
            return xml_item

        def DumpInfo(self) -> None:
            """Prints internal information to the console."""
            print("\t\t  {TypeString:s}".format(TypeString=self.TypeString))
            print("\t\t--------------------------------------------------")
            print("\t\tType                  : 0x{Type:02X}".format(Type=self.Type))
            print("\t\tStart of Range        : 0x{DeviceID:04X}".format(DeviceID=self.DeviceID))
            print("\t\tEnd of Range          : 0x{EndDeviceID:04X}".format(EndDeviceID=self.EndDeviceID))
            print("\t\tDTE Setting           : 0x{DTESetting:02X}".format(DTESetting=self.DTESetting))

    class DEVICE_TABLE_ENTRY_ALIAS_SELECT(object):
        """Object representing a Device Table Entry ALIAS_SELECT."""

        def __init__(self, data: Optional[bytes] = None) -> "IVRS_TABLE.DEVICE_TABLE_ENTRY_ALIAS_SELECT":
            """Inits an empty object."""
            self.Type = 0
            self.DeviceID = 0
            self.DTESetting = 0
            self.SourceDeviceID = 0
            self.TypeString = None
            self.Length = 0
            if data is not None:
                self.Decode(data)

        def Decode(self, header_byte_array: bytes) -> None:
            """Loads data into the object from the header_byte_array."""
            (self.Type, self.DeviceID, self.DTESetting) = struct.unpack(
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                header_byte_array[: IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size],
            )
            if self.Type != IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.ALIAS_SELECT:
                raise Exception(
                    "Input device type (%d) does not match expectation (%d)",
                    IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.ALIAS_SELECT,
                    self.Type,
                )

            self.TypeString = "Alias Select"
            # Two DevID, one for alias, one for source
            self.Length = (
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size + IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size
            )
            (_, self.SourceDeviceID, _) = struct.unpack(
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                header_byte_array[IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size : self.Length],
            )

        def Encode(self) -> bytes:
            r"""Serializes the object.

            Returns:
                (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
            """
            byte_str = struct.pack(
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format, self.Type, self.DeviceID, self.DTESetting
            )
            byte_str += struct.pack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format, 0, self.SourceDeviceID, 0)
            return byte_str

        def ToXmlElementTree(self) -> ET.Element:
            """Transforms the object into an xml item and returns it."""
            xml_item = ET.Element(self.TypeString.replace(" ", ""))
            xml_item.set("Type", "0x%X" % self.Type)
            xml_item.set("DeviceID", "0x%X" % (self.DeviceID))
            xml_item.set("DTESetting", "0x%X" % (self.DTESetting))
            xml_item.set("SourceDeviceID", "0x%X" % (self.SourceDeviceID))
            return xml_item

        def DumpInfo(self) -> None:
            """Prints internal information to the console."""
            print("\t\t  {TypeString:s}".format(TypeString=self.TypeString))
            print("\t\t--------------------------------------------------")
            print("\t\tType                  : 0x{Type:02X}".format(Type=self.Type))
            print("\t\tDevice ID             : 0x{DeviceID:04X}".format(DeviceID=self.DeviceID))
            print("\t\tDTE Setting           : 0x{DTESetting:02X}".format(DTESetting=self.DTESetting))
            print("\t\tSource Device ID      : 0x{SourceDeviceID:04X}".format(SourceDeviceID=self.SourceDeviceID))

    class DEVICE_TABLE_ENTRY_ALIAS_RANGE_START(object):
        """Object representing a Device Table Range Start."""

        def __init__(self, data: Optional[bytes] = None) -> "IVRS_TABLE.DEVICE_TABLE_ENTRY_ALIAS_RANGE_START":
            """Inits an empty object."""
            self.Type = 0
            self.DeviceID = 0
            self.DTESetting = 0
            self.SourceDeviceID = 0
            self.EndDeviceID = 0
            self.TypeString = None
            self.Length = 0
            if data is not None:
                self.Decode(data)

        def Decode(self, header_byte_array: bytes) -> None:
            """Loads data into the object from the header_byte_array."""
            (self.Type, self.DeviceID, self.DTESetting) = struct.unpack(
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                header_byte_array[: IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size],
            )
            if self.Type != IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.ALIAS_RANGE_START:
                raise Exception(
                    "Input device type (%d) does not match expectation (%d)",
                    IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.ALIAS_RANGE_START,
                    self.Type,
                )

            self.TypeString = "Alias Range"
            # Two DevID, one for alias start, one for source start
            self.Length = (
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size + IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size
            )
            (_, self.SourceDeviceID, _) = struct.unpack(
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                header_byte_array[IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size : self.Length],
            )
            (Type, self.EndDeviceID, _) = struct.unpack(
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                header_byte_array[self.Length : (self.Length + IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size)],
            )
            if Type != IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.RANGE_END:
                print("Start of range does not follow end of range")
                sys.exit(-1)
            self.Length += IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size

        def Encode(self) -> bytes:
            r"""Serializes the object.

            Returns:
                (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
            """
            byte_str = struct.pack(
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format, self.Type, self.DeviceID, self.DTESetting
            )
            byte_str += struct.pack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format, 0, self.SourceDeviceID, 0)
            byte_str += struct.pack(
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.RANGE_END,
                self.EndDeviceID,
                0,
            )
            return byte_str

        def ToXmlElementTree(self) -> ET.Element:
            """Transforms the object into an xml item and returns it."""
            xml_item = ET.Element(self.TypeString.replace(" ", ""))
            xml_item.set("Type", "0x%X" % self.Type)
            xml_item.set("StartofRange", "0x%X" % self.DeviceID)
            xml_item.set("EndofRange", "0x%X" % (self.EndDeviceID))
            xml_item.set("DTESetting", "0x%X" % (self.DTESetting))
            xml_item.set("SourceDeviceID", "0x%X" % (self.SourceDeviceID))
            return xml_item

        def DumpInfo(self) -> None:
            """Prints internal information to the console."""
            print("\t\t  {TypeString:s}".format(TypeString=self.TypeString))
            print("\t\t--------------------------------------------------")
            print("\t\tType                  : 0x{Type:02X}".format(Type=self.Type))
            print("\t\tStart of Range        : 0x{DeviceID:04X}".format(DeviceID=self.DeviceID))
            print("\t\tEnd of Range          : 0x{EndDeviceID:04X}".format(EndDeviceID=self.EndDeviceID))
            print("\t\tDTE Setting           : 0x{DTESetting:02X}".format(DTESetting=self.DTESetting))
            print("\t\tSource Device ID      : 0x{SourceDeviceID:04X}".format(SourceDeviceID=self.SourceDeviceID))

    class DEVICE_TABLE_ENTRY_EX_SELECT(object):
        """Object representing a Device Table Entry EX_SELECT."""

        def __init__(self, data: Optional[bytes] = None) -> "IVRS_TABLE.DEVICE_TABLE_ENTRY_EX_SELECT":
            """Inits an empty object."""
            self.Type = 0
            self.DeviceID = 0
            self.DTESetting = 0
            self.ExtendedDTESetting = 0
            self.TypeString = None
            self.Length = 0
            if data is not None:
                self.Decode(data)

        def Decode(self, header_byte_array: bytes) -> None:
            """Loads data into the object from the header_byte_array."""
            (self.Type, self.DeviceID, self.DTESetting) = struct.unpack(
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                header_byte_array[: IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size],
            )
            if self.Type != IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.EX_SELECT:
                raise Exception(
                    "Input device type (%d) does not match expectation (%d)",
                    IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.EX_SELECT,
                    self.Type,
                )

            self.TypeString = "Extended Select"
            # Two DTE setting, one for standard setting, one for extended setting (AtsDisabled, etc.)
            self.Length = (
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size + IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size
            )
            (self.ExtendedDTESetting,) = struct.unpack(
                "=I", header_byte_array[IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size : self.Length]
            )

        def Encode(self) -> bytes:
            r"""Serializes the object.

            Returns:
                (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
            """
            byte_str = struct.pack(
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format, self.Type, self.DeviceID, self.DTESetting
            )
            # Two DTE setting, one for standard setting, one for extended setting (AtsDisabled, etc.)
            byte_str += struct.pack("=I", self.ExtendedDTESetting)
            return byte_str

        def ToXmlElementTree(self) -> ET.Element:
            """Transforms the object into an xml item and returns it."""
            xml_item = ET.Element(self.TypeString.replace(" ", ""))
            xml_item.set("Type", "0x%X" % self.Type)
            xml_item.set("DeviceID", "0x%X" % (self.DeviceID))
            xml_item.set("DTESetting", "0x%X" % (self.DTESetting))
            if (self.ExtendedDTESetting & 0x80000000) != 0:
                xml_item.set("ExtendedDTESetting", "ATS requests blocked")
            else:
                xml_item.set("ExtendedDTESetting", "ATS allowed")
            return xml_item

        def DumpInfo(self) -> None:
            """Prints internal information to the console."""
            print("\t\t  {TypeString:s}".format(TypeString=self.TypeString))
            print("\t\t--------------------------------------------------")
            print("\t\tType                  : 0x{Type:02X}".format(Type=self.Type))
            print("\t\tDevice ID             : 0x{DeviceID:04X}".format(DeviceID=self.DeviceID))
            print("\t\tDTE Setting           : 0x{DTESetting:02X}".format(DTESetting=self.DTESetting))
            if (self.ExtendedDTESetting & 0x80000000) != 0:
                ats_str = "ATS requests blocked"
            else:
                ats_str = "ATS allowed"
            print("\t\tExtended DTE Setting  : {ExtendedDTESetting:s}".format(ExtendedDTESetting=ats_str))

    class DEVICE_TABLE_ENTRY_EX_RANGE_START(object):
        """Object representing a Device Table Entry EX_RANGE_START."""

        def __init__(self, data: Optional[bytes] = None) -> "IVRS_TABLE.DEVICE_TABLE_ENTRY_EX_RANGE_START":
            """Inits an empty object."""
            self.Type = 0
            self.DeviceID = 0
            self.DTESetting = 0
            self.TypeString = None
            self.Length = 0
            if data is not None:
                self.Decode(data)

        def Decode(self, header_byte_array: bytes) -> None:
            """Loads data into the object from the header_byte_array."""
            (self.Type, self.DeviceID, self.DTESetting) = struct.unpack(
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                header_byte_array[: IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size],
            )
            if self.Type != IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.EX_RANGE_START:
                raise Exception(
                    "Input device type (%d) does not match expectation (%d)",
                    IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.EX_RANGE_START,
                    self.Type,
                )

            self.TypeString = "Extended Range"
            # Two DTE setting, one for standard setting start, one for extended setting start
            self.Length = (
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size + IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size
            )
            (self.ExtendedDTESetting,) = struct.unpack(
                "=I", header_byte_array[IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size : self.Length]
            )
            (Type, self.EndDeviceID, _) = struct.unpack(
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                header_byte_array[self.Length : (self.Length + IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size)],
            )
            if Type != IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.RANGE_END:
                print("Start of range does not follow end of range")
                sys.exit(-1)
            self.Length += IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size

        def Encode(self) -> bytes:
            r"""Serializes the object.

            Returns:
                (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
            """
            byte_str = struct.pack(
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format, self.Type, self.DeviceID, self.DTESetting
            )
            # Two DTE setting, one for standard setting start, one for extended setting start
            byte_str += struct.pack("=I", self.ExtendedDTESetting)
            byte_str += struct.pack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format, 4, self.EndDeviceID, 0)
            return byte_str

        def ToXmlElementTree(self) -> ET.Element:
            """Transforms the object into an xml item and returns it."""
            xml_item = ET.Element(self.TypeString.replace(" ", ""))
            xml_item.set("Type", "0x%X" % self.Type)
            xml_item.set("StartofRange", "0x%X" % self.DeviceID)
            xml_item.set("EndofRange", "0x%X" % (self.EndDeviceID))
            xml_item.set("DTESetting", "0x%X" % (self.DTESetting))
            if (self.ExtendedDTESetting & 0x80000000) != 0:
                xml_item.set("ExtendedDTESetting", "ATS requests blocked")
            else:
                xml_item.set("ExtendedDTESetting", "ATS allowed")
            return xml_item

        def DumpInfo(self) -> None:
            """Prints internal information to the console."""
            print("\t\t  {TypeString:s}".format(TypeString=self.TypeString))
            print("\t\t--------------------------------------------------")
            print("\t\tType                  : 0x{Type:02X}".format(Type=self.Type))
            print("\t\tStart of Range        : 0x{DeviceID:04X}".format(DeviceID=self.DeviceID))
            print("\t\tEnd of Range          : 0x{EndDeviceID:04X}".format(EndDeviceID=self.EndDeviceID))
            print("\t\tDTE Setting           : 0x{DTESetting:02X}".format(DTESetting=self.DTESetting))
            if (self.ExtendedDTESetting & 0x80000000) != 0:
                ats_str = "ATS requests blocked"
            else:
                ats_str = "ATS allowed"
            print("\t\tExtended DTE Setting  : {ExtendedDTESetting:s}".format(ExtendedDTESetting=ats_str))

    class DEVICE_TABLE_ENTRY_SPECIAL(object):
        """Object representing a Device Table Entry Special."""

        def __init__(self, data: Optional[bytes] = None) -> "IVRS_TABLE.DEVICE_TABLE_ENTRY_SPECIAL":
            """Inits an empty object."""
            self.Type = 0
            self.DeviceID = 0
            self.DTESetting = 0
            self.ExtendedDTESetting = 0
            self.TypeString = None
            self.Length = 0
            if data is not None:
                self.Decode(data)

        def Decode(self, header_byte_array: bytes) -> None:
            """Loads data into the object from the header_byte_array."""
            (self.Type, self.DeviceID, self.DTESetting) = struct.unpack(
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                header_byte_array[: IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size],
            )
            if self.Type != IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.SPECIAL:
                raise Exception(
                    "Input device type (%d) does not match expectation (%d)",
                    IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.SPECIAL,
                    self.Type,
                )

            self.TypeString = "Special Device"
            # First half for standard DTE setting, second half for special DevID and its variety (APIC, HPET, etc.)
            self.Length = (
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size + IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size
            )
            (self.Handle, self.SourceDeviceID, self.Variety) = struct.unpack(
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                header_byte_array[IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size : self.Length],
            )

        def Encode(self) -> bytes:
            r"""Serializes the object.

            Returns:
                (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
            """
            byte_str = struct.pack(
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format, self.Type, self.DeviceID, self.DTESetting
            )
            # First half for standard DTE setting, second half for special DevID and its variety (APIC, HPET, etc.)
            byte_str += struct.pack(
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format, self.Handle, self.SourceDeviceID, self.Variety
            )
            return byte_str

        def ToXmlElementTree(self) -> ET.Element:
            """Transforms the object into an xml item and returns it."""
            xml_item = ET.Element(self.TypeString.replace(" ", ""))
            xml_item.set("Type", "0x%X" % self.Type)
            xml_item.set("DeviceID", "0x%X" % (self.DeviceID))
            xml_item.set("DTESetting", "0x%X" % (self.DTESetting))
            xml_item.set("SourceDeviceID", "0x%X" % (self.SourceDeviceID))

            xml_item.set("Handle", "0x%X" % (self.Handle))
            if self.Variety == 1:
                xml_item.set("Variety", "IOAPIC")
            elif self.Variety == 2:
                xml_item.set("Variety", "HPET")
            else:
                xml_item.set("Variety", "Reserved %X" % (self.Variety))
            return xml_item

        def DumpInfo(self) -> None:
            """Prints internal information to the console."""
            print("\t\t  {TypeString:s}".format(TypeString=self.TypeString))
            print("\t\t--------------------------------------------------")
            print("\t\tType                  : 0x{Type:02X}".format(Type=self.Type))
            print("\t\tDevice ID             : 0x{DeviceID:04X}".format(DeviceID=self.DeviceID))
            print("\t\tDTE Setting           : 0x{DTESetting:02X}".format(DTESetting=self.DTESetting))
            print("\t\tSource Device ID      : 0x{SourceDeviceID:04X}".format(SourceDeviceID=self.SourceDeviceID))

            if self.Variety == 1:
                var_str = "IOAPIC"
            elif self.Variety == 2:
                var_str = "HPET"
            else:
                var_str = "Reserved 0x%02X" % (self.Variety)
            print("\t\tHandle                : 0x{Handle:02X}".format(Handle=self.Handle))
            print("\t\tVariety               : {Variety:s}".format(Variety=var_str))

    class DEVICE_TABLE_ENTRY_ACPI(object):
        """Object representing a Device Table Entry ACPI."""

        def __init__(self, data: Optional[bytes] = None) -> "IVRS_TABLE.DEVICE_TABLE_ENTRY_ACPI":
            """Inits an empty object."""
            self.Type = 0
            self.DeviceID = 0
            self.DTESetting = 0
            self.ExtendedDTESetting = 0
            self.TypeString = None
            self.Length = 0
            if data is not None:
                self.Decode(data)

        def Decode(self, header_byte_array: bytes) -> None:
            """Loads data into the object from the header_byte_array."""
            (self.Type, self.DeviceID, self.DTESetting) = struct.unpack(
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                header_byte_array[: IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size],
            )
            if self.Type != IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.ACPI:
                raise Exception(
                    "Input device type (%d) does not match expectation (%d)",
                    IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.ACPI,
                    self.Type,
                )

            self.TypeString = "Variable Length ACPI HID Device"
            (self.HID, self.CID, self.UIDFormat, self.UIDLength) = struct.unpack(
                IVRS_TABLE.DEVICE_TABLE_ENTRY.dte_var_ext_format,
                header_byte_array[
                    IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size : IVRS_TABLE.DEVICE_TABLE_ENTRY.dte_var_len
                ],
            )
            self.Length = IVRS_TABLE.DEVICE_TABLE_ENTRY.dte_var_len + self.UIDLength
            if self.UIDFormat == 0:
                self.UID = None
            elif self.UIDFormat == 1:
                (self.UID,) = struct.unpack(
                    "=Q", header_byte_array[IVRS_TABLE.DEVICE_TABLE_ENTRY.dte_var_len : self.Length]
                )
            elif self.UIDFormat == 2:
                (self.UID,) = struct.unpack(
                    "=%ss" % self.UIDLength, header_byte_array[IVRS_TABLE.DEVICE_TABLE_ENTRY.dte_var_len : self.Length]
                )

        def Encode(self) -> bytes:
            r"""Serializes the object.

            Returns:
                (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
            """
            byte_str = struct.pack(
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format, self.Type, self.DeviceID, self.DTESetting
            )
            # Variable Length ACPI HID Device
            byte_str += struct.pack(
                IVRS_TABLE.DEVICE_TABLE_ENTRY.dte_var_ext_format, self.HID, self.CID, self.UIDFormat, self.UIDLength
            )
            if self.UIDFormat == 1:
                byte_str += struct.pack("=Q", self.UID)
            elif self.UIDFormat == 2:
                byte_str += struct.pack("=%ss" % self.UIDLength, self.UID)
            return byte_str

        def ToXmlElementTree(self) -> ET.Element:
            """Transforms the object into an xml item and returns it."""
            xml_item = ET.Element(self.TypeString.replace(" ", ""))
            xml_item.set("Type", "0x%X" % self.Type)
            xml_item.set("DeviceID", "0x%X" % (self.DeviceID))
            xml_item.set("DTESetting", "0x%X" % (self.DTESetting))

            xml_item.set("HardwareID", "%s" % (self.HID))
            xml_item.set("ExtendedDTE Setting", "%s" % (self.CID))
            xml_item.set("UniqueIDFormat", "%d" % (self.UIDFormat))
            xml_item.set("UniqueIDLength", "%d" % (self.UIDLength))
            if self.UIDFormat == 0:
                xml_item.set("UniqueID", "None")
            elif self.UIDFormat == 1:
                xml_item.set("UniqueID", "0x%X" % (self.UID))
            elif self.UIDFormat == 2:
                xml_item.set("UniqueID", "%s" % (self.UID))
            else:
                print("Unrecognized UID format detected")
                sys.exit(-1)
            return xml_item

        def DumpInfo(self) -> None:
            """Prints internal information to the console."""
            print("\t\t  {TypeString:s}".format(TypeString=self.TypeString))
            print("\t\t--------------------------------------------------")
            print("\t\tType                  : 0x{Type:02X}".format(Type=self.Type))
            print("\t\tDevice ID             : 0x{DeviceID:04X}".format(DeviceID=self.DeviceID))
            print("\t\tDTE Setting           : 0x{DTESetting:02X}".format(DTESetting=self.DTESetting))

            print("\t\tHardware ID           : {HID:s}".format(HID=self.HID.decode()))
            print("\t\tExtended DTE Setting  : {CID:s}".format(CID=self.CID.decode()))
            print("\t\tUnique ID Format      : {UIDFormat:d}".format(UIDFormat=self.UIDFormat))
            print("\t\tUnique ID Length      : {UIDLength:d}".format(UIDLength=self.UIDLength))
            if self.UIDFormat == 0:
                print("\t\tUnique ID             : None")
            elif self.UIDFormat == 1:
                print("\t\tUnique ID             : 0x{UID:X}".format(UID=self.UID))
            elif self.UIDFormat == 2:
                print("\t\tUnique ID             : {UID:s}".format(UID=self.UID.decode()))
            else:
                raise Exception("Unrecognized UID format detected %d" % self.UIDFormat)



# @file
# Module contains helper classes and functions to work with UEFI FVs.
#
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Module containing helper classes and functions for working with UEFI Fvs."""

import struct
import sys
import uuid
from typing import IO

#
# UEFI GUIDs
#
EfiSystemNvDataFvGuid = uuid.UUID(fields=(0xFFF12B8D, 0x7696, 0x4C8B, 0xA9, 0x85, 0x2747075B4F50))


#
# UEFI #Defines
#
EFI_FVH_SIGNATURE = b"_FVH"


class EfiFirmwareVolumeHeader(object):
    """An object representing an EFI_FIRMWARE_VOLUME_HEADER.

    Can parse or produce an EFI_FIRMWARE_VOLUME_HEADER structure/byte buffer.

    ```
    typedef struct {
        UINT8                     ZeroVector[16];
        EFI_GUID                  FileSystemGuid;
        UINT64                    FvLength;
        UINT32                    Signature;
        EFI_FVB_ATTRIBUTES_2      Attributes;
        UINT16                    HeaderLength;
        UINT16                    Checksum;
        UINT16                    ExtHeaderOffset;
        UINT8                     Reserved[1];
        UINT8                     Revision;
        EFI_FV_BLOCK_MAP_ENTRY    BlockMap[1];
    } EFI_FIRMWARE_VOLUME_HEADER;
    ```
    """

    def __init__(self) -> "EfiFirmwareVolumeHeader":
        """Inits an empty object."""
        self.StructString = "=16s16sQ4sLHHHBBQQ"  # spell-checker: disable-line
        self.ZeroVector = None
        self.FileSystemGuid = None
        self.FvLength = None
        self.Attributes = None
        self.HeaderLength = None
        self.Checksum = None
        self.ExtHeaderOffset = None
        self.Reserved = None
        self.Revision = None
        self.Blockmap0 = None
        self.Blockmap1 = None

    def load_from_file(self, file: IO) -> "EfiFirmwareVolumeHeader":
        """Loads data into the object from a filestream.

        Args:
            file (obj): An open file that has been seeked to the correct location.

        Returns:
            (EfiFirmwareVolumeHeader): self

        Raises:
            (Exception): Invalid signature in fs
        """
        # This function assumes that the file has been seeked
        # to the correct starting location.
        orig_seek = file.tell()
        struct_bytes = file.read(struct.calcsize(self.StructString))
        file.seek(orig_seek)

        # Load this object with the contents of the data.
        (
            self.ZeroVector,
            file_system_guid_bin,
            self.FvLength,
            self.Signature,
            self.Attributes,
            self.HeaderLength,
            self.Checksum,
            self.ExtHeaderOffset,
            self.Reserved,
            self.Revision,
            self.Blockmap0,
            self.Blockmap1,
        ) = struct.unpack(self.StructString, struct_bytes)

        # Make sure that this structure is what we think it is.
        if self.Signature != EFI_FVH_SIGNATURE:
            raise Exception("File does not appear to point to a valid EfiFirmwareVolumeHeader!")

        # Update the GUID to be a UUID object.
        if sys.byteorder == "big":
            self.FileSystemGuid = uuid.UUID(bytes=file_system_guid_bin)
        else:
            self.FileSystemGuid = uuid.UUID(bytes_le=file_system_guid_bin)

        return self

    def serialize(self) -> bytes:
        r"""Serializes the object.

        Returns:
            (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
        """
        file_system_guid_bin = self.FileSystemGuid.bytes if sys.byteorder == "big" else self.FileSystemGuid.bytes_le
        return struct.pack(
            self.StructString,
            self.ZeroVector,
            file_system_guid_bin,
            self.FvLength,
            self.Signature,
            self.Attributes,
            self.HeaderLength,
            self.Checksum,
            self.ExtHeaderOffset,
            self.Reserved,
            self.Revision,
            self.Blockmap0,
            self.Blockmap1,
        )


class EfiFirmwareVolumeExtHeader(object):
    """An object representing an EFI_FIRMWARE_VOLUME_EXT_HEADER.

    Can parse or produce an EFI_FIRMWARE_VOLUME_EXT_HEADER structure/byte buffer.

    ```
    typedef struct {
        EFI_GUID                  FileSystemGuid;
        UINT32                    ExtHeaderSize;
    } EFI_FIRMWARE_VOLUME_EXT_HEADER;
    ```
    """

    def __init__(self) -> "EfiFirmwareVolumeExtHeader":
        """Inits an empty object."""
        self.StructString = "=16sL"
        self.FileSystemGuid = None
        self.ExtHeaderSize = None

    def load_from_file(self, file: IO) -> "EfiFirmwareVolumeExtHeader":
        """Loads data into the object from a filestream.

        Args:
            file: An open file that has been seeked to the correct location.

        Returns:
            (EfiFirmwareVolumeExtHeader): self
        """
        # This function assumes that the file has been seeked
        # to the correct starting location.
        orig_seek = file.tell()
        struct_bytes = file.read(struct.calcsize(self.StructString))
        file.seek(orig_seek)

        # Load this object with the contents of the data.
        (self.FileSystemGuid, self.ExtHeaderSize) = struct.unpack(self.StructString, struct_bytes)

        return self



# @file fdf_parser.py
# Code to help parse EDK2 Fdf files
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Code to help parse EDK2 Fdf files."""

import os

from edk2toollib.uefi.edk2.parsers.base_parser import HashFileParser


class FdfParser(HashFileParser):
    """FDF parser object that represents an FDF file.

    Attributes:
        Parsed (bool): Whether the object contains a parsed FDF file
        Lines (list[str]): an ordered list of each line in the FDF file
        Dict (dict): dict of form: {"VariableName": Value}
        FVs (dict): dict of form: {"sectionName": {"Dict": {}, "Infs": [], "Files": {}}}
        FDs (dict): dict of form: {"sectionName": {}}
        CurrentSection (list): all lines in current section
        Path (str): path to FDF file

    Note: Dict Key Value pairs come from lines that contain a single =.
    """

    def __init__(self) -> "HashFileParser":
        """Inits an empty FDF parser."""
        HashFileParser.__init__(self, "ModuleFdfParser")
        self.Lines = []
        self.Parsed = False
        self.Dict = {}  # defines dictionary
        self.FVs = {}
        self.FDs = {}
        self.CurrentSection = []
        self.Path = ""

    def GetNextLine(self) -> str:
        """Returns the next line to parse.

        Performs manipulation on the line like replacing variables,
        processing conditionals, etc.
        """
        if len(self.Lines) == 0:
            return None

        line = self.Lines.pop()
        self.CurrentLine += 1
        sline = self.StripComment(line)

        if sline is None or len(sline) < 1:
            return self.GetNextLine()

        sline = self.ReplaceVariables(sline)
        if self.ProcessConditional(sline):
            # was a conditional so skip
            return self.GetNextLine()
        if not self.InActiveCode():
            return self.GetNextLine()

        self._BracketCount += sline.count("{")
        self._BracketCount -= sline.count("}")

        return sline

    def InsertLinesFromFile(self, file_path: str) -> None:
        """Adds additional lines to the Lines Attribute from the provided file."""
        with open(file_path, "r") as lines_file:
            self.Lines += reversed(lines_file.readlines())
            # Back off the line count to ignore the include line itself.
            self.CurrentLine -= 1

    def ParseFile(self, filepath: str) -> None:
        """Parses the provided FDF file."""
        self.Logger.debug("Parsing file: %s" % filepath)
        if not os.path.isabs(filepath):
            fp = self.FindPath(filepath)
        else:
            fp = filepath
        self.Path = fp
        self.TargetFilePath = os.path.abspath(fp)
        self.CurrentLine = 0
        self._f = open(fp, "r")
        self.Lines = self._f.readlines()
        self.Lines.reverse()
        self._f.close()
        self._BracketCount = 0
        InDefinesSection = False
        InFdSection = False
        InFvSection = False
        InCapsuleSection = False
        InFmpPayloadSection = False
        InRuleSection = False

        sline = ""
        while sline is not None:
            sline = self.GetNextLine()

            if sline is None:
                break

            if sline.lower().startswith("!include"):
                tokens = sline.split()
                include_file = tokens[1]
                sp = self.FindPath(include_file)
                if sp is None:
                    raise FileNotFoundError(include_file)
                self.Logger.debug("Opening Include File %s" % sp)
                self.InsertLinesFromFile(sp)
                continue

            if sline.strip().startswith("[") and sline.strip().endswith("]"):  # if we're starting a new section
                # this basically gets what's after the . or if it doesn't have a period
                # the whole thing for every comma separated item in sline
                self.CurrentSection = [
                    x.split(".", 1)[1] if "." in x else x for x in sline.strip("[] ").strip().split(",")
                ]
                InDefinesSection = False
                InFdSection = False
                InFvSection = False
                InCapsuleSection = False
                InFmpPayloadSection = False
                InRuleSection = False
                self.LocalVars = {}
                self.LocalVars.update(self.Dict)

            if InDefinesSection:
                if sline.count("=") == 1:
                    tokens = sline.replace("DEFINE", "").split("=", 1)
                    self.Dict[tokens[0].strip()] = tokens[1].strip()
                    self.Logger.info("Key,values found:  %s = %s" % (tokens[0].strip(), tokens[1].strip()))
                    continue

            # defining a local variable that is removed when entering a new section
            elif sline.strip().startswith("DEFINE"):
                tokens = sline.strip().replace("DEFINE", "").split("=", 1)
                self.LocalVars[tokens[0].strip()] = tokens[1].strip()
                self.Logger.info(f"Key,values found for local vars: {tokens[0].strip()}, {tokens[1].strip()}")
                continue

            elif InFdSection:
                for section in self.CurrentSection:
                    if section not in self.FVs:
                        self.FDs[section] = {"Dict": {}}
                        # TODO finish the FD section
                continue

            elif InFvSection:
                for section in self.CurrentSection:
                    if section not in self.FVs:
                        self.FVs[section] = {"Dict": {}, "Infs": [], "Files": {}}
                    # ex: INF  MdeModulePkg/Core/RuntimeDxe/RuntimeDxe.inf
                    if sline.upper().startswith("INF "):
                        InfValue = sline[3:].strip()
                        self.FVs[section]["Infs"].append(InfValue)
                    # ex: FILE FREEFORM = 7E175642-F3AD-490A-9F8A-2E9FC6933DDD {
                    elif sline.upper().startswith("FILE"):
                        sline = sline.strip("}").strip("{").strip()  # make sure we take off the { and }
                        file_def = sline[4:].strip().split("=", 1)  # split by =
                        if len(file_def) != 2:  # check to make sure we can parse this file
                            raise RuntimeError("Unable to properly parse " + sline)

                        currentType = file_def[0].strip()  # get the type FILE
                        currentName = file_def[1].strip()  # get the name (guid or otherwise)
                        if currentType not in self.FVs[section]:
                            self.FVs[section]["Files"][currentName] = {}
                        self.FVs[section]["Files"][currentName]["type"] = currentType

                        while self._BracketCount > 0:  # go until we get our bracket back
                            sline = self.GetNextLine().strip("}{ ")
                            # SECTION GUIDED EE4E5898-3914-4259-9D6E-DC7BD79403CF
                            if sline.upper().startswith("SECTION GUIDED"):  # get the guided section
                                section_def = sline[14:].strip().split("=", 1)
                                # EE4E5898-3914-4259-9D6E-DC7BD79403CF in this example
                                sectionType = section_def[0].strip()
                                if sectionType not in self.FVs[section]["Files"][currentName]:
                                    self.FVs[section]["Files"][currentName][sectionType] = {}
                                # TODO support guided sections
                            # ex: SECTION UI = "GenericGopDriver"
                            elif (
                                sline.upper().startswith("SECTION") and sline.upper().count("=") > 0
                            ):  # get the section
                                section_def = sline[7:].strip().split("=", 1)
                                sectionType = section_def[0].strip()  # UI in this example
                                sectionValue = section_def[1].strip()

                                if sectionType not in self.FVs[section]["Files"][currentName]:
                                    self.FVs[section]["Files"][currentName][sectionType] = []
                                self.FVs[section]["Files"][currentName][sectionType].append(sectionValue)
                            else:
                                self.Logger.info("Unknown line: {}".format(sline))

                continue

            elif InCapsuleSection:
                # TODO: finish capsule section
                continue

            elif InFmpPayloadSection:
                # TODO finish FMP payload section
                continue

            elif InRuleSection:
                # TODO finish rule section
                continue

            # check for different sections
            if sline.strip().lower().startswith("[defines"):
                InDefinesSection = True

            elif sline.strip().lower().startswith("[fd."):
                InFdSection = True

            elif sline.strip().lower().startswith("[fv."):
                InFvSection = True

            elif sline.strip().lower().startswith("[capsule."):
                InCapsuleSection = True

            elif sline.strip().lower().startswith("[fmpPayload."):
                InFmpPayloadSection = True

            elif sline.strip().lower().startswith("[rule."):
                InRuleSection = True

        self.Parsed = True



# @file source_table.py
# A module to Parse all Source files and add them to the database.
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""A module to Parse all Source files and add them to the database."""

import logging
import re
import time
from pathlib import Path
from typing import Any

from joblib import Parallel, delayed
from pygount import SourceAnalysis

from edk2toollib.database import Session, Source
from edk2toollib.database.tables import TableGenerator
from edk2toollib.uefi.edk2.path_utilities import Edk2Path

SOURCE_EXT_LIST = ["*.c", "*.h", "*.cpp", "*.asm", "*.s", "*.nasm", "*.masm", "*.rs"]


class SourceTable(TableGenerator):
    """A Table Generator that parses all c and h files in the workspace."""

    def __init__(self, *args: Any, **kwargs: Any) -> "SourceTable":
        """Initializes the Source Table Parser.

        Args:
            args (any): non-keyword arguments
            kwargs (any): keyword arguments described below

        Keyword Arguments:
            source_stats (bool): Whether to parse source statistics
            n_jobs (int): Number of files to run in parallel
            source_extensions (list[str]): List of file extensions to parse
        """
        self.source_stats = kwargs.get("source_stats", False)
        self.n_jobs = kwargs.get("n_jobs", -1)
        self.source_extensions = kwargs.get("source_extensions", SOURCE_EXT_LIST)

    def parse(self, session: Session, pathobj: Edk2Path, id: str, env: dict) -> None:
        """Parse the workspace and update the database."""
        ws = Path(pathobj.WorkspacePath)
        self.pathobj = pathobj

        start = time.time()
        files = []
        for src in self.source_extensions:
            files.extend(list(ws.rglob(src)))
        files = [file for file in files if not file.is_relative_to(ws / "Build")]
        src_entries = Parallel(n_jobs=self.n_jobs)(delayed(self._parse_file)(filename) for filename in files)

        existing_source = {source.path: source for source in session.query(Source).all()}
        to_add = []
        for source in src_entries:
            if source.path not in existing_source:
                existing_source[source.path] = source
                to_add.append(source)
            else:
                to_add.append(existing_source[source.path])
                to_add[-1].code_lines = source.code_lines
                to_add[-1].comment_lines = source.comment_lines
                to_add[-1].blank_lines = source.blank_lines

        session.add_all(to_add)
        session.commit()

        logging.debug(
            f"{self.__class__.__name__}: Parsed {len(src_entries)} files; "
            f"took {round(time.time() - start, 2)} seconds."
        )

    def _parse_file(self, filename: Path) -> dict:
        """Parse a C file and return the results."""
        license = ""
        with open(filename, "r", encoding="cp850") as f:
            lines = f.readlines()
            for line in lines:
                match = re.search(r"SPDX-License-Identifier:\s*(.*)$", line)  # TODO: This is not a standard format.
                if match:
                    license = match.group(1)

        total_lines = len(lines)
        code_lines = total_lines
        comment_lines = 0
        blank_lines = 0
        if self.source_stats:
            code = SourceAnalysis.from_file(filename, "_", fallback_encoding="utf-8")
            code_lines = code.code_count
            comment_lines = code.documentation_count
            blank_lines = code.empty_count

        path = self.pathobj.GetEdk2RelativePathFromAbsolutePath(filename.as_posix())
        return Source(
            path=path,
            license=license or "Unknown",
            total_lines=total_lines,
            code_lines=code_lines,
            comment_lines=comment_lines,
            blank_lines=blank_lines,
        )



# @file guid_parser.py
# Code to help parse guid formats and make into UUIDs
#
# Some functionality copied from Tianocore/edk2 basetools
#
# Copyright (c) Microsoft Corporation
# Copyright (c) 2007 - 2019, Intel Corporation. All rights reserved.<BR>
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Code to help parse guid formats and transform into UUIDs.

Some functionality copied from Tianocore/edk2 basetools.
"""

import re
import uuid


class GuidParser:
    """Provide support functions for converting between different guid formats.

    Also support str uuid and uuid to string.

    Note:
      C-Format:   {0xD3B36F2C, 0xD551, 0x11D4, {0x9A, 0x46, 0x00, 0x90, 0x27, 0x3F, 0xC1, 0x4D}}
      Reg-Format: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
    """

    _HexChar = r"[0-9a-fA-F]"
    # Regular expression for GUID c structure format
    _GuidCFormatPattern = (
        r"{{\s*0[xX]{Hex}{{1,8}}\s*,\s*0[xX]{Hex}{{1,4}}\s*,\s*0[xX]{Hex}{{1,4}}"
        r"\s*,\s*{{\s*0[xX]{Hex}{{1,2}}\s*,\s*0[xX]{Hex}{{1,2}}"
        r"\s*,\s*0[xX]{Hex}{{1,2}}\s*,\s*0[xX]{Hex}{{1,2}}"
        r"\s*,\s*0[xX]{Hex}{{1,2}}\s*,\s*0[xX]{Hex}{{1,2}}"
        r"\s*,\s*0[xX]{Hex}{{1,2}}\s*,\s*0[xX]{Hex}{{1,2}}\s*}}\s*}}".format(Hex=_HexChar)
    )
    GuidCFormatRegEx = re.compile(r"{}".format(_GuidCFormatPattern))

    _GuidPattern = r"{Hex}{{8}}-{Hex}{{4}}-{Hex}{{4}}-{Hex}{{4}}-{Hex}{{12}}".format(Hex=_HexChar)

    # Regular expressions for GUID matching
    GuidRegFormatRegEx = re.compile(r"{}".format(_GuidPattern))

    @classmethod
    def is_guid_in_c_format(cls: "GuidParser", guidstring: str) -> bool:
        """Determine if guidstring is in c format.

        Args:
          guidstring (str): string containing guid

        Returns:
          (bool): True if in C format. Otherwise False

        """
        guidstring = guidstring.strip()
        return cls.GuidCFormatRegEx.match(guidstring)

    @classmethod
    def is_guid_in_reg_format(cls: "GuidParser", guidstring: str) -> bool:
        """Determine if guidstring is in registry format.

        Args:
          guidstring (str): string containing guid

        Returns:
          (bool): True if in Registry format. Otherwise False
        """
        guidstring = guidstring.strip().strip("} {")
        return cls.GuidRegFormatRegEx.match(guidstring)

    @classmethod
    def reg_guid_from_c_format(cls: "GuidParser", guidstring: str) -> str:
        """Convert a c formatted guidstring to a registry formatted guidstring.

        Args:
          guidstring (str): c format guidstring

        Returns:
          (Success): guidstring in registry format
          (Failure): empty string ''
        """
        guidstring = guidstring.strip()
        if not cls.is_guid_in_c_format(guidstring):
            return ""

        guidValueString = guidstring.lower().replace("{", "").replace("}", "").replace(" ", "").replace(";", "")
        guidValueList = guidValueString.split(",")
        if len(guidValueList) != 11:
            return ""
        try:
            return "%08x-%04x-%04x-%02x%02x-%02x%02x%02x%02x%02x%02x" % (
                int(guidValueList[0], 16),
                int(guidValueList[1], 16),
                int(guidValueList[2], 16),
                int(guidValueList[3], 16),
                int(guidValueList[4], 16),
                int(guidValueList[5], 16),
                int(guidValueList[6], 16),
                int(guidValueList[7], 16),
                int(guidValueList[8], 16),
                int(guidValueList[9], 16),
                int(guidValueList[10], 16),
            )
        except Exception:
            return ""

    @classmethod
    def c_guid_from_reg_format(cls: "GuidParser", guidstring: str) -> str:
        """Convert registry format guidstring to c format guidstring.

        Args:
          guidstring (str): registry format guidstring

        Returns:
          (Success): guidstring in c format
          (Failure): empty string ''
        """
        guidstring = guidstring.strip().strip("} {")
        if not cls.is_guid_in_reg_format(guidstring):
            return ""

        GuidList = guidstring.split("-")
        Result = "{"
        for Index in range(0, 3, 1):
            Result = Result + "0x" + GuidList[Index] + ", "
        Result = Result + "{0x" + GuidList[3][0:2] + ", 0x" + GuidList[3][2:4]
        for Index in range(0, 12, 2):
            Result = Result + ", 0x" + GuidList[4][Index : Index + 2]
        Result += "}}"
        return Result

    @classmethod
    def uuid_from_guidstring(cls: "GuidParser", guidstring: str) -> uuid.UUID:
        """Create a uuid object from the supplied guidstring."""
        if cls.is_guid_in_c_format(guidstring):
            return uuid.UUID(cls.reg_guid_from_c_format(guidstring))
        elif cls.is_guid_in_reg_format(guidstring):
            guidstring = guidstring.strip().strip("} {")
            return uuid.UUID(guidstring)
        else:
            return None

    @classmethod
    def c_guid_str_from_uuid(cls: "GuidParser", guid: uuid.UUID) -> str:
        """Get a C string formatted guidstring from a uuid object.

        Args:
          guid (uuid.UUID): valid uuid object

        Returns:
          (Success): guidstring in C format
          (Failure): empty string ''
        """
        reg = str(guid)
        return cls.c_guid_from_reg_format(reg)

    @classmethod
    def reg_guid_str_from_uuid(cls: "GuidParser", guid: uuid.UUID) -> str:
        """Get a registry string formatted guidstring from a uuid object.

        Args:
          guid (uuid.UUID): valid uuid object

        Returns:
          (Success): guidstring in registry format
          (Failure): empty string ''
        """
        return str(guid)



##
# Python script that converts a raw DMAR table into a struct
# More details see https://software.intel.com/sites/default/files/managed/c5/15/vt-directed-io-spec.pdf
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Module for converting a raw DMAR table into a struct."""

import os
import struct
import sys
import xml.etree.ElementTree as ET
from typing import Optional

DMARParserVersion = "1.01"


class DMARTable(object):
    """Object representing the DMAR Table."""

    # Header Lengths
    DMARHeaderLength = 48
    DRDHHeaderLength = 16
    RMRRHeaderLength = 24
    ASTRHeaderLength = 8
    ANDDHeaderLength = 8
    DeviceScopeHeaderLength = 6

    def __init__(self, data: bytes) -> "DMARTable":
        """Inits the object."""
        self.dmar_table = self.AcpiTableHeader(data)
        self.data = data[DMARTable.DMARHeaderLength :]
        while len(self.data) > 0:
            # Get type and length of remapping struct
            remapping_header = self.RemappingStructHeader(self.data)
            assert remapping_header.Type < 5, "Reserved remapping struct found in DMAR table"

            # Parse remapping struct
            if remapping_header.Type == 0:
                remapping_header = self.DRHDStruct(self.data, remapping_header.Length)
            elif remapping_header.Type == 1:
                remapping_header = self.RMRRStruct(self.data, remapping_header.Length)
                self.dmar_table.RMRRlist.append(remapping_header)
            elif remapping_header.Type == 2:
                remapping_header = self.ATSRStruct(self.data, remapping_header.Length)
            elif remapping_header.Type == 3:
                remapping_header = self.RHSAStruct(self.data, remapping_header.Length)
            elif remapping_header.Type == 4:
                remapping_header = self.ANDDStruct(self.data, remapping_header.Length)
                self.dmar_table.ANDDCount += 1
            else:
                print("Reserved remapping struct found in DMAR table")
                sys.exit(-1)

            self.dmar_table.SubStructs.append(remapping_header)
            # Add to XML
            self.data = self.data[remapping_header.Length :]

        self.xml = self.toXml()

    def toXml(self) -> ET.Element:
        """Converts the object to an xml representation."""
        root = ET.Element("DMAR Table")
        root.append(self.dmar_table.toXml())
        for sub in self.dmar_table.SubStructs:
            root.append(sub.toXml())

        return root

    def __str__(self) -> str:
        """String representation of the object."""
        retval = str(self.dmar_table)

        for sub in self.dmar_table.SubStructs:
            retval += str(sub)

        return retval

    def DMARBitEnabled(self) -> bool:
        """Returns the status of the DMAR Bit."""
        return bool(self.dmar_table.DMARBit)

    def ANDDCount(self) -> int:
        """Returns the amount of ANDD in the DMAR table."""
        return self.dmar_table.ANDDCount

    def CheckRMRRCount(self, goldenxml: Optional[ET.Element] = None) -> bool:
        """Verifies all RMRR paths are in the golden XML."""
        goldenignores = set()

        if goldenxml is None or not os.path.isfile(goldenxml):
            print("XML File not found")
        else:
            goldenfile = ET.parse(goldenxml)
            goldenroot = goldenfile.getroot()
            for RMRR in goldenroot:
                goldenignores.add(RMRR.find("Path").text.lower())

        for RMRR in self.dmar_table.RMRRlist:
            if RMRR.getPath() not in goldenignores:
                print("RMRR PCIe Endpoint " + RMRR.getPath() + " found but not in golden XML")
                return False

        return True

    class AcpiTableHeader(object):
        """Object representing the ACPI Table Header."""

        STRUCT_FORMAT = "=4sIBB6s8sI4sIBB"
        size = struct.calcsize(STRUCT_FORMAT)

        def __init__(self, header_byte_array: bytes) -> "DMARTable.AcpiTableHeader":
            """Inits the object."""
            (
                self.Signature,
                self.Length,
                self.Revision,
                self.Checksum,
                self.OEMID,
                self.OEMTableID,
                self.OEMRevision,
                self.CreatorID,
                self.CreatorRevision,
                self.HostAddressWidth,
                self.Flags,
            ) = struct.unpack_from(DMARTable.AcpiTableHeader.STRUCT_FORMAT, header_byte_array)

            self.DMARBit = self.Flags & 0x4
            self.ANDDCount = 0
            self.RMRRlist = list()
            self.SubStructs = list()

        def __str__(self) -> str:
            """String representation of the object."""
            return """\n  ACPI Table Header
    ------------------------------------------------------------------
      Signature          : %s
      Length             : 0x%08X
      Revision           : 0x%02X
      Checksum           : 0x%02X
      OEM ID             : %s
      OEM Table ID       : %s
      OEM Revision       : 0x%08X
      Creator ID         : %s
      Creator Revision   : 0x%08X
      Host Address Width : 0x%02X
      Flags              : 0x%02X\n""" % (
                self.Signature,
                self.Length,
                self.Revision,
                self.Checksum,
                self.OEMID,
                self.OEMTableID,
                self.OEMRevision,
                self.CreatorID,
                self.CreatorRevision,
                self.HostAddressWidth,
                self.Flags,
            )

        def toXml(self) -> ET.Element:
            """Converts the object to an xml representation."""
            xml_repr = ET.Element("AcpiTableHeader")
            xml_repr.set("Signature", "%s" % self.Signature)
            xml_repr.set("Length", "0x%X" % self.Length)
            xml_repr.set("Revision", "0x%X" % self.Revision)
            xml_repr.set("Checksum", "0x%X" % self.Checksum)
            xml_repr.set("OEMID", "%s" % self.OEMID)
            xml_repr.set("OEMTableID", "%s" % self.OEMTableID)
            xml_repr.set("OEMRevision", "0x%X" % self.OEMRevision)
            xml_repr.set("CreatorID", "%s" % self.CreatorID)
            xml_repr.set("CreatorRevision", "0x%X" % self.CreatorRevision)
            xml_repr.set("HostAddressWidth", "0x%X" % self.HostAddressWidth)
            xml_repr.set("Flags", "0x%X" % self.Flags)
            return xml_repr

    class RemappingStructHeader(object):
        """Generic remapping struct header."""

        STRUCT_FORMAT = "=HH"

        def __init__(self, header_byte_array: bytes) -> "DMARTable.RemappingStructHeader":
            """Inits the object."""
            (self.Type, self.Length) = struct.unpack_from(
                DMARTable.RemappingStructHeader.STRUCT_FORMAT, header_byte_array
            )

        def __str__(self) -> str:
            """String representation of the object."""
            return """\n  Remapping Struct Header
    ------------------------------------------------------------------
      Type               : 0x%04X
      Length             : 0x%04X
    """ % (self.Type, self.Length)

    class DRHDStruct(RemappingStructHeader):
        """Object representing the DRHD struct."""

        STRUCT_FORMAT = "=HHBBHQ"  # spell-checker: disable-line

        def __init__(self, header_byte_array: bytes, length: int) -> "DMARTable.DRHDStruct":
            """Inits the object."""
            (self.Type, self.Length, self.Flags, self.Reserved, self.SegmentNumber, self.RegisterBaseAddress) = (
                struct.unpack_from(DMARTable.DRHDStruct.STRUCT_FORMAT, header_byte_array)
            )

            # Get Sub Structs
            self.DeviceScope = list()
            header_byte_array = header_byte_array[DMARTable.DRDHHeaderLength :]
            bytes_left = self.Length - DMARTable.DRDHHeaderLength
            while bytes_left > 0:
                device_scope = DMARTable.DeviceScopeStruct(header_byte_array)
                header_byte_array = header_byte_array[device_scope.Length :]
                bytes_left -= device_scope.Length
                self.DeviceScope.append(device_scope)

        def toXml(self) -> ET.Element:
            """Converts the object to an xml representation."""
            xml_repr = ET.Element("DRHD")
            xml_repr.set("Type", "0x%X" % self.Type)
            xml_repr.set("Length", "0x%X" % self.Length)
            xml_repr.set("Flags", "0x%X" % self.Flags)
            xml_repr.set("Reserved", "0x%X" % self.Reserved)
            xml_repr.set("SegmentNumber", "0x%X" % self.SegmentNumber)
            xml_repr.set("RegisterBaseAddress", "0x%X" % self.RegisterBaseAddress)

            # Add SubStructs
            for item in self.DeviceScope:
                xml_subitem = ET.SubElement(xml_repr, item.TypeString)
                xml_subitem.set("Type", "0x%X" % item.Type)
                xml_subitem.set("Length", "0x%X" % item.Length)
                xml_subitem.set("Reserved", "0x%X" % item.Reserved)
                xml_subitem.set("EnumerationID", "0x%X" % item.EnumerationID)
                xml_subitem.set("StartBusNumber", "0x%X" % item.StartBusNumber)

            return xml_repr

        def __str__(self) -> str:
            """String representation of the object."""
            retstring = """\n  DRHD
    ------------------------------------------------------------------
      Type                  : 0x%04X
      Length                : 0x%04X
      Flags                 : 0x%02X
      Reserved              : 0x%02X
      Segment Number        : 0x%04x
      Register Base Address : 0x%016x
    """ % (self.Type, self.Length, self.Flags, self.Reserved, self.SegmentNumber, self.RegisterBaseAddress)

            for item in self.DeviceScope:
                retstring += str(item)

            return retstring

    class RMRRStruct(RemappingStructHeader):
        """Object representing the RMRR struct."""

        STRUCT_FORMAT = "=HHHHQQ"  # spell-checker: disable-line

        def __init__(self, header_byte_array: bytes, length: int) -> "DMARTable.RMRRStruct":
            """Inits the object."""
            (
                self.Type,
                self.Length,
                self.Reserved,
                self.SegmentNumber,
                self.ReservedMemoryBaseAddress,
                self.ReservedMemoryRegionLimitAddress,
            ) = struct.unpack_from(DMARTable.RMRRStruct.STRUCT_FORMAT, header_byte_array)

            # Get Sub Structs
            self.DeviceScope = list()
            header_byte_array = header_byte_array[DMARTable.RMRRHeaderLength :]
            bytes_left = self.Length - DMARTable.RMRRHeaderLength
            while bytes_left > 0:
                device_scope = DMARTable.DeviceScopeStruct(header_byte_array)
                header_byte_array = header_byte_array[device_scope.Length :]
                bytes_left -= device_scope.Length
                self.DeviceScope.append(device_scope)

        def getPath(self) -> str:
            """Generates and returns the path."""
            retString = ""
            for index, item in enumerate(self.DeviceScope):
                retString += self.DeviceScope[index].getPath()
                if index != len(self.DeviceScope) - 1:
                    retString += ", "
            return retString

        def toXml(self) -> ET.Element:
            """Converts the object to an xml representation."""
            xml_repr = ET.Element("RMRR")
            xml_repr.set("Type", "0x%X" % self.Type)
            xml_repr.set("Length", "0x%X" % self.Length)
            xml_repr.set("Reserved", "0x%X" % self.Reserved)
            xml_repr.set("SegmentNumber", "0x%X" % self.SegmentNumber)
            xml_repr.set("ReservedMemoryBaseAddress", "0x%X" % self.ReservedMemoryBaseAddress)
            xml_repr.set("ReservedMemoryRegionLimitAddress", "0x%X" % self.ReservedMemoryRegionLimitAddress)

            # Add SubStructs
            for item in self.DeviceScope:
                xml_subitem = ET.SubElement(xml_repr, item.TypeString)
                xml_subitem.set("Type", "0x%X" % item.Type)
                xml_subitem.set("Length", "0x%X" % item.Length)
                xml_subitem.set("Reserved", "0x%X" % item.Reserved)
                xml_subitem.set("EnumerationID", "0x%X" % item.EnumerationID)
                xml_subitem.set("StartBusNumber", "0x%X" % item.StartBusNumber)

            return xml_repr

        def __str__(self) -> str:
            """String representation of the object."""
            retstring = """\n  RMRR
    ------------------------------------------------------------------
      Type                                 : 0x%04X
      Length                               : 0x%04X
      Reserved                             : 0x%04X
      Segment Number                       : 0x%04x
      Reserved Memory Base Address         : 0x%016x
      Reserved Memory Region Limit Address : 0x%016x\n""" % (
                self.Type,
                self.Length,
                self.Reserved,
                self.SegmentNumber,
                self.ReservedMemoryBaseAddress,
                self.ReservedMemoryRegionLimitAddress,
            )

            for item in self.DeviceScope:
                retstring += str(item)

            return retstring

    class ATSRStruct(RemappingStructHeader):
        """Object representing the ANDD struct."""

        STRUCT_FORMAT = "=HHBBH"  # spell-checker: disable-line

        def __init__(self, header_byte_array: bytes, length: int) -> "DMARTable.ATSRStruct":
            """Inits the object."""
            (self.Type, self.Length, self.Flags, self.Reserved, self.SegmentNumber) = struct.unpack_from(
                DMARTable.ATSRStruct.STRUCT_FORMAT, header_byte_array
            )

            # Get Sub Structs
            self.DeviceScope = list()
            header_byte_array = header_byte_array[DMARTable.ASTRHeaderLength :]
            bytes_left = self.Length - DMARTable.ASTRHeaderLength
            while bytes_left > 0:
                device_scope = DMARTable.DeviceScopeStruct(header_byte_array)
                header_byte_array = header_byte_array[device_scope.Length :]
                bytes_left -= device_scope.Length
                self.DeviceScope.append(device_scope)

        def toXml(self) -> ET.Element:
            """Converts the object to an xml representation."""
            xml_repr = ET.Element("ASTR")
            xml_repr.set("Type", "0x%X" % self.Type)
            xml_repr.set("Length", "0x%X" % self.Length)
            xml_repr.set("Flags", "0x%X" % self.Flags)
            xml_repr.set("Reserved", "0x%X" % self.Reserved)
            xml_repr.set("SegmentNumber", "0x%X" % self.SegmentNumber)

            # Add SubStructs
            for item in self.DeviceScope:
                xml_subitem = ET.SubElement(xml_repr, item.TypeString)
                xml_subitem.set("Type", "0x%X" % item.Type)
                xml_subitem.set("Length", "0x%X" % item.Length)
                xml_subitem.set("Reserved", "0x%X" % item.Reserved)
                xml_subitem.set("EnumerationID", "0x%X" % item.EnumerationID)
                xml_subitem.set("StartBusNumber", "0x%X" % item.StartBusNumber)

            return xml_repr

        def __str__(self) -> str:
            """String representation of the object."""
            retstring = """\n  ASTR
    ------------------------------------------------------------------
      Type                                 : 0x%04X
      Length                               : 0x%04X
      Flags                                : 0x%02X
      Reserved                             : 0x%02X
      Segment Number                       : 0x%04x
    """ % (self.Type, self.Length, self.Flags, self.Reserved, self.SegmentNumber)

            for item in self.DeviceScope:
                retstring += str(item)

            return retstring

    class RHSAStruct(RemappingStructHeader):
        """Object representing the RHSA struct."""

        STRUCT_FORMAT = "=HHIQI"  # spell-checker: disable-line

        def __init__(self, header_byte_array: bytes, length: int) -> "DMARTable.RHSAStruct":
            """Inits the object."""
            (self.Type, self.Length, self.Reserved, self.RegisterBaseAddress, self.ProximityDomain) = (
                struct.unpack_from(DMARTable.RHSAStruct.STRUCT_FORMAT, header_byte_array)
            )

        def toXml(self) -> ET.Element:
            """Converts the object to an xml representation."""
            xml_repr = ET.Element("RHSA")
            xml_repr.set("Type", "0x%X" % self.Type)
            xml_repr.set("Length", "0x%X" % self.Length)
            xml_repr.set("Reserved", "0x%X" % self.Reserved)
            xml_repr.set("RegisterBaseAddress", "0x%X" % self.RegisterBaseAddress)
            xml_repr.set("ProximityDomain", "0x%X" % self.ProximityDomain)

            return xml_repr

        def __str__(self) -> str:
            """String representation of the object."""
            return """\n  RHSA
    ------------------------------------------------------------------
      Type                                 : 0x%04X
      Length                               : 0x%04X
      Reserved                             : 0x%08X
      Register Base Address                : 0x%016X
      Proximity Domain                     : 0x%08x
    """ % (self.Type, self.Length, self.Reserved, self.RegisterBaseAddress, self.ProximityDomain)

    class ANDDStruct(RemappingStructHeader):
        """Object representing the ANDD struct."""

        header_format = "=HH"

        def __init__(self, header_byte_array: bytes, length: int) -> "DMARTable.ANDDStruct":
            """Inits the object."""
            self.STRUCT_FORMAT = "=B"
            (self.Type, self.Length) = struct.unpack_from(DMARTable.ANDDStruct.header_format, header_byte_array)

            # Since there is no variable of size 3 we need to manually pull into reserved
            self.Reserved = 0
            for i in range(6, 3, -1):
                self.Reserved = self.Reserved << 8
                self.Reserved |= struct.unpack("<B", header_byte_array[i : i + 1])[0]
            header_byte_array = header_byte_array[7:]

            # Unpack remaining values
            self.STRUCT_FORMAT = self.STRUCT_FORMAT + str(self.Length - DMARTable.ANDDHeaderLength) + "s"
            (self.ACPIDeviceNumber, self.ACPIObjectName) = struct.unpack_from(self.STRUCT_FORMAT, header_byte_array)

        def toXml(self) -> ET.Element:
            """Converts the object to an xml representation."""
            xml_repr = ET.Element("ANDD")
            xml_repr.set("Type", "0x%X" % self.Type)
            xml_repr.set("Length", "0x%X" % self.Length)
            xml_repr.set("Reserved", "0x%X" % self.Reserved)
            xml_repr.set("ACPIDeviceNumber", "0x%X" % self.ACPIDeviceNumber)
            xml_repr.set("ACPIObjectName", "%s" % self.ACPIObjectName)

            return xml_repr

        def __str__(self) -> str:
            """String representation of the object."""
            return """\n  ANDD
    ------------------------------------------------------------------
      Type                                 : 0x%04X
      Length                               : 0x%04X
      Reserved                             : 0x%06X
      ACPI Device Number                   : 0x%02X
      ACPI Object Name                     : %s
    """ % (self.Type, self.Length, self.Reserved, self.ACPIDeviceNumber, self.ACPIObjectName)

    class DeviceScopeStruct(object):
        """Object representing a Device Scope."""

        STRUCT_FORMAT = "=BBHBB"  # spell-checker: disable-line

        def __init__(self, header_byte_array: bytes) -> "DMARTable.DeviceScopeStruct":
            """Inits a DeviceScopeStruct."""
            (self.Type, self.Length, self.Reserved, self.EnumerationID, self.StartBusNumber) = struct.unpack_from(
                DMARTable.DeviceScopeStruct.STRUCT_FORMAT, header_byte_array
            )

            assert self.Type < 6, "Reserved Device Scope Type Found"

            if self.Type == 1:
                self.TypeString = "PCI Endpoint Device"
            elif self.Type == 2:
                self.TypeString = "PCI Sub-hierarchy"
            elif self.Type == 3:
                self.TypeString = "IOAPIC"
            elif self.Type == 4:
                self.TypeString = "MSI_CAPABLE_HPET"
            elif self.Type == 5:
                self.TypeString = "ACPI_NAMESPACE_DEVICE"
            else:
                print("Reserved Device Scope Type Found")
                sys.exit(-1)

            number_path_entries = (self.Length - DMARTable.DeviceScopeHeaderLength) / 2
            offset = 6
            self.Path = list()
            while number_path_entries > 0:
                self.Path.append(
                    (
                        struct.unpack("<B", header_byte_array[offset : offset + 1]),
                        struct.unpack("<B", header_byte_array[offset + 1 : offset + 2]),
                    )
                )
                offset += 2
                number_path_entries -= 1

        def getPath(self) -> str:
            """Returns the path."""
            retstring = "%02d" % self.StartBusNumber + ":"

            for index, item in enumerate(self.Path):
                retstring += "%02d" % item[0] + "." + "%01d" % item[1]
                if index != len(self.Path) - 1:
                    retstring += ":"

            return retstring

        def __str__(self) -> str:
            """String representation."""
            retstring = """\n\t\t  %s
    \t\t--------------------------------------------------
    \t\t  Type                  : 0x%02X
    \t\t  Length                : 0x%02X
    \t\t  Reserved              : 0x%04X
    \t\t  Enumeration ID        : 0x%02x
    \t\t  Start Bus Number      : 0x%02x
    \t\t  Path                  : """ % (
                self.TypeString,
                self.Type,
                self.Length,
                self.Reserved,
                self.EnumerationID,
                self.StartBusNumber,
            )

            retstring += "%02d" % self.StartBusNumber + ":"
            for index, item in enumerate(self.Path):
                retstring += "%02d" % item[0] + "." + "%01d" % item[1]
                if index != len(self.Path) - 1:
                    retstring += ":"
            retstring += "\n"

            return retstring



##
# Handle basic logging by streaming into stringIO
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Module for handling basic logging by streaming into StringIO."""

import io
import logging
from logging import LogRecord


class StringStreamHandler(logging.StreamHandler):
    """Class for logging via StringIO."""

    terminator = "\n"

    def __init__(self) -> "StringStreamHandler":
        """Init a StringStreamHandler."""
        logging.Handler.__init__(self)
        self.stream = io.StringIO()

    def handle(self, record: LogRecord) -> bool:
        """Conditionally emit the specified logging record.

        Emission depends on filters which may have been added to the handler.
        Wrap the actual emission of the record with acquisition/release of
        the I/O thread lock. Returns whether the filter passed the record for
        emission.
        """
        rv = self.filter(record)
        if rv and record.levelno >= self.level:
            self.acquire()
            try:
                self.emit(record)
            finally:
                self.release()
        return rv

    def readlines(self, hint: int = -1) -> list[str]:
        """Reads lines from stream and returns them."""
        return self.stream.readlines(hint)

    def seek_start(self) -> None:
        """Seeks to a specific point in the stream."""
        self.stream.seek(0, 0)

    def seek_end(self) -> None:
        """Seeks to the end of the stream."""
        self.stream.seek(0, io.SEEK_END)

    def seek(self, offset: int, whence: int) -> int:
        """Seek to a specific point in the stream."""
        return self.stream.seek(offset, whence)
